@@ -14,6 +14,8 @@ import (
 	"path/filepath"
 
 	"golang.org/x/tools/internal/event"
+	"golang.org/x/tools/internal/gocommand"
+	"golang.org/x/tools/internal/lsp/cache"
 	"golang.org/x/tools/internal/lsp/protocol"
 	"golang.org/x/tools/internal/lsp/source"
 	"golang.org/x/tools/internal/span"
@@ -24,7 +26,7 @@ import (
 func (s *Server) executeCommand(ctx context.Context, params *protocol.ExecuteCommandParams) (interface{}, error) {
 	var command *source.Command
 	for _, c := range source.Commands {
-		if c.Name == params.Command {
+		if c.ID() == params.Command {
 			command = c
 			break
 		}
@@ -34,17 +36,13 @@ func (s *Server) executeCommand(ctx context.Context, params *protocol.ExecuteCom
 	}
 	var match bool
 	for _, name := range s.session.Options().SupportedCommands {
-		if command.Name == name {
+		if command.ID() == name {
 			match = true
 			break
 		}
 	}
 	if !match {
-		return nil, fmt.Errorf("%s is not a supported command", command.Name)
-	}
-	title := command.Title
-	if title == "" {
-		title = command.Name
+		return nil, fmt.Errorf("%s is not a supported command", command.ID())
 	}
 	// Some commands require that all files are saved to disk. If we detect
 	// unsaved files, warn the user instead of running the commands.
@@ -57,48 +55,39 @@ func (s *Server) executeCommand(ctx context.Context, params *protocol.ExecuteCom
 	}
 	if unsaved {
 		switch params.Command {
-		case source.CommandTest.Name, source.CommandGenerate.Name, source.CommandToggleDetails.Name:
+		case source.CommandTest.ID(),
+			source.CommandGenerate.ID(),
+			source.CommandToggleDetails.ID(),
+			source.CommandAddDependency.ID(),
+			source.CommandUpgradeDependency.ID(),
+			source.CommandRemoveDependency.ID(),
+			source.CommandVendor.ID():
 			// TODO(PJW): for Toggle, not an error if it is being disabled
-			err := errors.New("unsaved files in the view")
-			s.showCommandError(ctx, title, err)
+			err := errors.New("all files must be saved first")
+			s.showCommandError(ctx, command.Title, err)
 			return nil, err
 		}
 	}
-	// If the command has a suggested fix function available, use it and apply
-	// the edits to the workspace.
-	if command.IsSuggestedFix() {
-		err := s.runSuggestedFixCommand(ctx, command, params.Arguments)
-		if err != nil {
-			s.showCommandError(ctx, title, err)
-		}
-		return nil, err
-	}
 	ctx, cancel := context.WithCancel(xcontext.Detach(ctx))
-	// Start progress prior to spinning off a goroutine specifically so that
-	// clients are aware of the work item before the command completes. This
-	// matters for regtests, where having a continuous thread of work is
-	// convenient for assertions.
-	work := s.progress.start(ctx, title, "Running...", params.WorkDoneToken, cancel)
-	if command.Synchronous {
-		return nil, s.runCommand(ctx, work, command, params.Arguments)
+
+	var work *workDone
+	// Don't show progress for suggested fixes. They should be quick.
+	if !command.IsSuggestedFix() {
+		// Start progress prior to spinning off a goroutine specifically so that
+		// clients are aware of the work item before the command completes. This
+		// matters for regtests, where having a continuous thread of work is
+		// convenient for assertions.
+		work = s.progress.start(ctx, command.Title, "Running...", params.WorkDoneToken, cancel)
 	}
-	go func() {
-		defer cancel()
-		err := s.runCommand(ctx, work, command, params.Arguments)
-		switch {
-		case errors.Is(err, context.Canceled):
-			work.end(title + ": canceled")
-		case err != nil:
-			event.Error(ctx, fmt.Sprintf("%s: command error", title), err)
-			work.end(title + ": failed")
-			// Show a message when work completes with error, because the progress end
-			// message is typically dismissed immediately by LSP clients.
-			s.showCommandError(ctx, title, err)
-		default:
-			work.end(command.Name + ": completed")
-		}
-	}()
-	return nil, nil
+	if command.Async {
+		go func() {
+			defer cancel()
+			s.runCommand(ctx, work, command, params.Arguments)
+		}()
+		return nil, nil
+	}
+	defer cancel()
+	return nil, s.runCommand(ctx, work, command, params.Arguments)
 }
 
 func (s *Server) runSuggestedFixCommand(ctx context.Context, command *source.Command, args []json.RawMessage) error {
@@ -141,7 +130,26 @@ func (s *Server) showCommandError(ctx context.Context, title string, err error)
 	}
 }
 
-func (s *Server) runCommand(ctx context.Context, work *workDone, command *source.Command, args []json.RawMessage) error {
+func (s *Server) runCommand(ctx context.Context, work *workDone, command *source.Command, args []json.RawMessage) (err error) {
+	defer func() {
+		switch {
+		case errors.Is(err, context.Canceled):
+			work.end(command.Title + ": canceled")
+		case err != nil:
+			event.Error(ctx, fmt.Sprintf("%s: command error", command.Title), err)
+			work.end(command.Title + ": failed")
+			// Show a message when work completes with error, because the progress end
+			// message is typically dismissed immediately by LSP clients.
+			s.showCommandError(ctx, command.Title, err)
+		default:
+			work.end(command.ID() + ": completed")
+		}
+	}()
+	// If the command has a suggested fix function available, use it and apply
+	// the edits to the workspace.
+	if command.IsSuggestedFix() {
+		return s.runSuggestedFixCommand(ctx, command, args)
+	}
 	switch command {
 	case source.CommandTest:
 		var uri protocol.DocumentURI
@@ -188,14 +196,24 @@ func (s *Server) runCommand(ctx context.Context, work *workDone, command *source
 		if command == source.CommandVendor {
 			a = "vendor"
 		}
-		return s.directGoModCommand(ctx, uri, "mod", []string{a}...)
-	case source.CommandUpgradeDependency:
+		return s.directGoModCommand(ctx, uri, "mod", a)
+	case source.CommandAddDependency, source.CommandUpgradeDependency, source.CommandRemoveDependency:
 		var uri protocol.DocumentURI
 		var goCmdArgs []string
-		if err := source.UnmarshalArgs(args, &uri, &goCmdArgs); err != nil {
+		var addRequire bool
+		if err := source.UnmarshalArgs(args, &uri, &addRequire, &goCmdArgs); err != nil {
 			return err
 		}
-		return s.directGoModCommand(ctx, uri, "get", goCmdArgs...)
+		if addRequire {
+			// Using go get to create a new dependency results in an
+			// `// indirect` comment we may not want. The only way to avoid it
+			// is to add the require as direct first. Then we can use go get to
+			// update go.sum and tidy up.
+			if err := s.directGoModCommand(ctx, uri, "mod", append([]string{"edit", "-require"}, goCmdArgs...)...); err != nil {
+				return err
+			}
+		}
+		return s.directGoModCommand(ctx, uri, "get", append([]string{"-d"}, goCmdArgs...)...)
 	case source.CommandToggleDetails:
 		var fileURI span.URI
 		if err := source.UnmarshalArgs(args, &fileURI); err != nil {
@@ -203,10 +221,10 @@ func (s *Server) runCommand(ctx context.Context, work *workDone, command *source
 		}
 		pkgDir := span.URIFromPath(filepath.Dir(fileURI.Filename()))
 		s.gcOptimizationDetailsMu.Lock()
-		if _, ok := s.gcOptimizatonDetails[pkgDir]; ok {
-			delete(s.gcOptimizatonDetails, pkgDir)
+		if _, ok := s.gcOptimizationDetails[pkgDir]; ok {
+			delete(s.gcOptimizationDetails, pkgDir)
 		} else {
-			s.gcOptimizatonDetails[pkgDir] = struct{}{}
+			s.gcOptimizationDetails[pkgDir] = struct{}{}
 		}
 		s.gcOptimizationDetailsMu.Unlock()
 		// need to recompute diagnostics.
@@ -217,7 +235,7 @@ func (s *Server) runCommand(ctx context.Context, work *workDone, command *source
 		}
 		snapshot, release := sv.Snapshot(ctx)
 		defer release()
-		s.diagnoseSnapshot(snapshot)
+		s.diagnoseSnapshot(snapshot, nil, false)
 	case source.CommandGenerateGoplsMod:
 		var v source.View
 		if len(args) == 0 {
@@ -239,7 +257,7 @@ func (s *Server) runCommand(ctx context.Context, work *workDone, command *source
 		}
 		snapshot, release := v.Snapshot(ctx)
 		defer release()
-		modFile, err := snapshot.BuildWorkspaceModFile(ctx)
+		modFile, err := cache.BuildGoplsMod(ctx, v.Folder(), snapshot)
 		if err != nil {
 			return errors.Errorf("getting workspace mod file: %w", err)
 		}
@@ -252,7 +270,7 @@ func (s *Server) runCommand(ctx context.Context, work *workDone, command *source
 			return errors.Errorf("writing mod file: %w", err)
 		}
 	default:
-		return fmt.Errorf("unsupported command: %s", command.Name)
+		return fmt.Errorf("unsupported command: %s", command.ID())
 	}
 	return nil
 }
@@ -264,7 +282,12 @@ func (s *Server) directGoModCommand(ctx context.Context, uri protocol.DocumentUR
 	}
 	snapshot, release := view.Snapshot(ctx)
 	defer release()
-	return snapshot.RunGoCommandDirect(ctx, verb, args)
+	_, err = snapshot.RunGoCommandDirect(ctx, source.UpdateUserModFile, &gocommand.Invocation{
+		Verb:       verb,
+		Args:       args,
+		WorkingDir: filepath.Dir(uri.SpanURI().Filename()),
+	})
+	return err
 }
 
 func (s *Server) runTests(ctx context.Context, snapshot source.Snapshot, uri protocol.DocumentURI, work *workDone, tests, benchmarks []string) error {
@@ -285,8 +308,12 @@ func (s *Server) runTests(ctx context.Context, snapshot source.Snapshot, uri pro
 	// Run `go test -run Func` on each test.
 	var failedTests int
 	for _, funcName := range tests {
-		args := []string{pkgPath, "-v", "-count=1", "-run", fmt.Sprintf("^%s$", funcName)}
-		if err := snapshot.RunGoCommandPiped(ctx, "test", args, out, out); err != nil {
+		inv := &gocommand.Invocation{
+			Verb:       "test",
+			Args:       []string{pkgPath, "-v", "-count=1", "-run", fmt.Sprintf("^%s$", funcName)},
+			WorkingDir: filepath.Dir(uri.SpanURI().Filename()),
+		}
+		if err := snapshot.RunGoCommandPiped(ctx, source.Normal, inv, out, out); err != nil {
 			if errors.Is(err, context.Canceled) {
 				return err
 			}
@@ -297,8 +324,12 @@ func (s *Server) runTests(ctx context.Context, snapshot source.Snapshot, uri pro
 	// Run `go test -run=^$ -bench Func` on each test.
 	var failedBenchmarks int
 	for _, funcName := range benchmarks {
-		args := []string{pkgPath, "-v", "-run=^$", "-bench", fmt.Sprintf("^%s$", funcName)}
-		if err := snapshot.RunGoCommandPiped(ctx, "test", args, out, out); err != nil {
+		inv := &gocommand.Invocation{
+			Verb:       "test",
+			Args:       []string{pkgPath, "-v", "-run=^$", "-bench", fmt.Sprintf("^%s$", funcName)},
+			WorkingDir: filepath.Dir(uri.SpanURI().Filename()),
+		}
+		if err := snapshot.RunGoCommandPiped(ctx, source.Normal, inv, out, out); err != nil {
 			if errors.Is(err, context.Canceled) {
 				return err
 			}
@@ -336,21 +367,24 @@ func (s *Server) runTests(ctx context.Context, snapshot source.Snapshot, uri pro
 	})
 }
 
-func (s *Server) runGoGenerate(ctx context.Context, snapshot source.Snapshot, uri span.URI, recursive bool, work *workDone) error {
+func (s *Server) runGoGenerate(ctx context.Context, snapshot source.Snapshot, dir span.URI, recursive bool, work *workDone) error {
 	ctx, cancel := context.WithCancel(ctx)
 	defer cancel()
 
 	er := &eventWriter{ctx: ctx, operation: "generate"}
-	args := []string{"-x"}
-	dir := uri.Filename()
+
+	pattern := "."
 	if recursive {
-		dir = filepath.Join(dir, "...")
+		pattern = "..."
 	}
-	args = append(args, dir)
 
+	inv := &gocommand.Invocation{
+		Verb:       "generate",
+		Args:       []string{"-x", pattern},
+		WorkingDir: dir.Filename(),
+	}
 	stderr := io.MultiWriter(er, workDoneWriter{work})
-
-	if err := snapshot.RunGoCommandPiped(ctx, "generate", args, er, stderr); err != nil {
+	if err := snapshot.RunGoCommandPiped(ctx, source.Normal, inv, er, stderr); err != nil {
 		return err
 	}
 	return nil
