@@ -8,6 +8,7 @@ import (
 	"bufio"
 	"context"
 	"fmt"
+	"os"
 	"path/filepath"
 	"regexp"
 	"strings"
@@ -46,6 +47,7 @@ type buffer struct {
 	version int
 	path    string
 	content []string
+	dirty   bool
 }
 
 func (b buffer) text() string {
@@ -58,7 +60,8 @@ func (b buffer) text() string {
 //
 // The zero value for EditorConfig should correspond to its defaults.
 type EditorConfig struct {
-	Env map[string]string
+	Env        map[string]string
+	BuildFlags []string
 
 	// CodeLens is a map defining whether codelens are enabled, keyed by the
 	// codeLens command. CodeLens which are not present in this map are left in
@@ -84,6 +87,14 @@ type EditorConfig struct {
 
 	// EnableStaticcheck enables staticcheck analyzers.
 	EnableStaticcheck bool
+
+	// AllExperiments sets the "allExperiments" configuration, which enables
+	// all of gopls's opt-in settings.
+	AllExperiments bool
+
+	// Whether to send the current process ID, for testing data that is joined to
+	// the PID. This can only be set by one test.
+	SendPID bool
 }
 
 // NewEditor Creates a new Editor.
@@ -181,6 +192,10 @@ func (e *Editor) configuration() map[string]interface{} {
 		"completionBudget":        "10s",
 	}
 
+	if e.Config.BuildFlags != nil {
+		config["buildFlags"] = e.Config.BuildFlags
+	}
+
 	if e.Config.CodeLens != nil {
 		config["codelens"] = e.Config.CodeLens
 	}
@@ -193,9 +208,15 @@ func (e *Editor) configuration() map[string]interface{} {
 	if e.Config.EnableStaticcheck {
 		config["staticcheck"] = true
 	}
-	// Default to using the experimental workspace module mode.
-	config["experimentalWorkspaceModule"] = true
+	if e.Config.AllExperiments {
+		config["allExperiments"] = true
+	}
 
+	// TODO(rFindley): uncomment this if/when diagnostics delay is on by
+	// default... and probably change to the new settings name.
+	// config["experimentalDiagnosticsDelay"] = "10ms"
+
+	// ExperimentalWorkspaceModule is only set as a mode, not a configuration.
 	return config
 }
 
@@ -206,7 +227,7 @@ func (e *Editor) initialize(ctx context.Context, withoutWorkspaceFolders bool, e
 	if !withoutWorkspaceFolders {
 		rootURI := e.sandbox.Workdir.RootURI()
 		if editorRootPath != "" {
-			rootURI = toURI(e.sandbox.Workdir.filePath(editorRootPath))
+			rootURI = toURI(e.sandbox.Workdir.AbsPath(editorRootPath))
 		}
 		params.WorkspaceFolders = []protocol.WorkspaceFolder{{
 			URI:  string(rootURI),
@@ -217,6 +238,9 @@ func (e *Editor) initialize(ctx context.Context, withoutWorkspaceFolders bool, e
 	params.Capabilities.Window.WorkDoneProgress = true
 	// TODO: set client capabilities
 	params.InitializationOptions = e.configuration()
+	if e.Config.SendPID {
+		params.ProcessID = float64(os.Getpid())
+	}
 
 	// This is a bit of a hack, since the fake editor doesn't actually support
 	// watching changed files that match a specific glob pattern. However, the
@@ -247,10 +271,28 @@ func (e *Editor) onFileChanges(ctx context.Context, evts []FileEvent) {
 	if e.Server == nil {
 		return
 	}
+	e.mu.Lock()
 	var lspevts []protocol.FileEvent
 	for _, evt := range evts {
+		// Always send an on-disk change, even for events that seem useless
+		// because they're shadowed by an open buffer.
 		lspevts = append(lspevts, evt.ProtocolEvent)
+
+		if buf, ok := e.buffers[evt.Path]; ok {
+			// Following VS Code, don't honor deletions or changes to dirty buffers.
+			if buf.dirty || evt.ProtocolEvent.Type == protocol.Deleted {
+				continue
+			}
+
+			content, err := e.sandbox.Workdir.ReadFile(evt.Path)
+			if err != nil {
+				continue // A race with some other operation.
+			}
+			// During shutdown, this call will fail. Ignore the error.
+			_ = e.setBufferContentLocked(ctx, evt.Path, false, strings.Split(content, "\n"), nil)
+		}
 	}
+	e.mu.Unlock()
 	e.Server.DidChangeWatchedFiles(ctx, &protocol.DidChangeWatchedFilesParams{
 		Changes: lspevts,
 	})
@@ -262,34 +304,7 @@ func (e *Editor) OpenFile(ctx context.Context, path string) error {
 	if err != nil {
 		return err
 	}
-	return e.OpenFileWithContent(ctx, path, content)
-}
-
-// OpenFileWithContent creates a buffer for the given workdir-relative file
-// with the given contents.
-func (e *Editor) OpenFileWithContent(ctx context.Context, path, content string) error {
-	buf := newBuffer(path, content)
-	e.mu.Lock()
-	e.buffers[path] = buf
-	item := textDocumentItem(e.sandbox.Workdir, buf)
-	e.mu.Unlock()
-
-	if e.Server != nil {
-		if err := e.Server.DidOpen(ctx, &protocol.DidOpenTextDocumentParams{
-			TextDocument: item,
-		}); err != nil {
-			return errors.Errorf("DidOpen: %w", err)
-		}
-	}
-	return nil
-}
-
-func newBuffer(path, content string) buffer {
-	return buffer{
-		version: 1,
-		path:    path,
-		content: strings.Split(content, "\n"),
-	}
+	return e.createBuffer(ctx, path, false, content)
 }
 
 func textDocumentItem(wd *Workdir, buf buffer) protocol.TextDocumentItem {
@@ -310,7 +325,16 @@ func textDocumentItem(wd *Workdir, buf buffer) protocol.TextDocumentItem {
 // CreateBuffer creates a new unsaved buffer corresponding to the workdir path,
 // containing the given textual content.
 func (e *Editor) CreateBuffer(ctx context.Context, path, content string) error {
-	buf := newBuffer(path, content)
+	return e.createBuffer(ctx, path, true, content)
+}
+
+func (e *Editor) createBuffer(ctx context.Context, path string, dirty bool, content string) error {
+	buf := buffer{
+		version: 1,
+		path:    path,
+		content: strings.Split(content, "\n"),
+		dirty:   dirty,
+	}
 	e.mu.Lock()
 	e.buffers[path] = buf
 	item := textDocumentItem(e.sandbox.Workdir, buf)
@@ -392,6 +416,12 @@ func (e *Editor) SaveBufferWithoutActions(ctx context.Context, path string) erro
 	if err := e.sandbox.Workdir.WriteFile(ctx, path, content); err != nil {
 		return errors.Errorf("writing %q: %w", path, err)
 	}
+
+	e.mu.Lock()
+	buf.dirty = false
+	e.buffers[path] = buf
+	e.mu.Unlock()
+
 	if e.Server != nil {
 		params := &protocol.DidSaveTextDocumentParams{
 			TextDocument: protocol.VersionedTextDocumentIdentifier{
@@ -526,6 +556,13 @@ func (e *Editor) EditBuffer(ctx context.Context, path string, edits []Edit) erro
 	return e.editBufferLocked(ctx, path, edits)
 }
 
+func (e *Editor) SetBufferContent(ctx context.Context, path, content string) error {
+	e.mu.Lock()
+	defer e.mu.Unlock()
+	lines := strings.Split(content, "\n")
+	return e.setBufferContentLocked(ctx, path, true, lines, nil)
+}
+
 // BufferText returns the content of the buffer with the given name.
 func (e *Editor) BufferText(name string) string {
 	e.mu.Lock()
@@ -546,24 +583,29 @@ func (e *Editor) editBufferLocked(ctx context.Context, path string, edits []Edit
 	if !ok {
 		return fmt.Errorf("unknown buffer %q", path)
 	}
-	var (
-		content = make([]string, len(buf.content))
-		err     error
-		evts    []protocol.TextDocumentContentChangeEvent
-	)
+	content := make([]string, len(buf.content))
 	copy(content, buf.content)
-	content, err = editContent(content, edits)
+	content, err := editContent(content, edits)
 	if err != nil {
 		return err
 	}
+	return e.setBufferContentLocked(ctx, path, true, content, edits)
+}
 
+func (e *Editor) setBufferContentLocked(ctx context.Context, path string, dirty bool, content []string, fromEdits []Edit) error {
+	buf, ok := e.buffers[path]
+	if !ok {
+		return fmt.Errorf("unknown buffer %q", path)
+	}
 	buf.content = content
 	buf.version++
+	buf.dirty = dirty
 	e.buffers[path] = buf
 	// A simple heuristic: if there is only one edit, send it incrementally.
 	// Otherwise, send the entire content.
-	if len(edits) == 1 {
-		evts = append(evts, edits[0].toProtocolChangeEvent())
+	var evts []protocol.TextDocumentContentChangeEvent
+	if len(fromEdits) == 1 {
+		evts = append(evts, fromEdits[0].toProtocolChangeEvent())
 	} else {
 		evts = append(evts, protocol.TextDocumentContentChangeEvent{
 			Text: buf.text(),
@@ -698,7 +740,7 @@ func (e *Editor) codeAction(ctx context.Context, path string, rng *protocol.Rang
 		// Execute any commands. The specification says that commands are
 		// executed after edits are applied.
 		if action.Command != nil {
-			if _, err := e.Server.ExecuteCommand(ctx, &protocol.ExecuteCommandParams{
+			if _, err := e.ExecuteCommand(ctx, &protocol.ExecuteCommandParams{
 				Command:   action.Command.Command,
 				Arguments: action.Command.Arguments,
 			}); err != nil {
@@ -709,6 +751,33 @@ func (e *Editor) codeAction(ctx context.Context, path string, rng *protocol.Rang
 	return nil
 }
 
+func (e *Editor) ExecuteCommand(ctx context.Context, params *protocol.ExecuteCommandParams) (interface{}, error) {
+	if e.Server == nil {
+		return nil, nil
+	}
+	var match bool
+	// Ensure that this command was actually listed as a supported command.
+	for _, command := range e.serverCapabilities.ExecuteCommandProvider.Commands {
+		if command == params.Command {
+			match = true
+			break
+		}
+	}
+	if !match {
+		return nil, fmt.Errorf("unsupported command %q", params.Command)
+	}
+	result, err := e.Server.ExecuteCommand(ctx, params)
+	if err != nil {
+		return nil, err
+	}
+	// Some commands use the go command, which writes directly to disk.
+	// For convenience, check for those changes.
+	if err := e.sandbox.Workdir.CheckForFileChanges(ctx); err != nil {
+		return nil, err
+	}
+	return result, nil
+}
+
 func convertEdits(protocolEdits []protocol.TextEdit) []Edit {
 	var edits []Edit
 	for _, lspEdit := range protocolEdits {
@@ -761,16 +830,16 @@ func (e *Editor) RunGenerate(ctx context.Context, dir string) error {
 	if e.Server == nil {
 		return nil
 	}
-	absDir := e.sandbox.Workdir.filePath(dir)
+	absDir := e.sandbox.Workdir.AbsPath(dir)
 	jsonArgs, err := source.MarshalArgs(span.URIFromPath(absDir), false)
 	if err != nil {
 		return err
 	}
 	params := &protocol.ExecuteCommandParams{
-		Command:   source.CommandGenerate.Name,
+		Command:   source.CommandGenerate.ID(),
 		Arguments: jsonArgs,
 	}
-	if _, err := e.Server.ExecuteCommand(ctx, params); err != nil {
+	if _, err := e.ExecuteCommand(ctx, params); err != nil {
 		return fmt.Errorf("running generate: %v", err)
 	}
 	// Unfortunately we can't simply poll the workdir for file changes here,
